#include "cam.h"
#include "iprinter.h"
#include "utils.h"
#include "record.h"

#include "VmbC/VmbCommonTypes.h"
#include "VmbC/VmbCTypeDefinitions.h"
#include "VmbCPP/SharedPointerDefines.h"

#include <memory> //shared_ptr
#include <string>
#include <unistd.h>

#include <QDebug>

using State = Cam::State;
using namespace VmbCPP;
using utils::DELIM;

Cam::Cam(QString name, QString ip) : IPrinter(),
_vcam(nullptr),
_name(name),
_ip(ip),
_id(""),
_state(disconnected),
_rec()
{
	// qDebug() << __LINE__ << "-" << __PRETTY_FUNCTION__ << "";
}

Cam::~Cam()
{
	qDebug() << __FUNCTION__ << "():" << __LINE__ << "id:" << _id << ", usecount:" << _vcam.use_count();
	if( _state == recording )
		rec()->stop();

	// close() should be called automatically, when the last shared_ptr to the camera is destroyed
}

RecordPtr Cam::rec()
{
	if( !_rec )
		error(utils::errRecordInvalid);

	return _rec;
}

QString Cam::name() const
{
	return _name;
}

QString Cam::ip() const
{
	return _ip;
}

QString Cam::id()
{
	if( _id.isEmpty() )
	{
		std::string str;
		f(_vcam->GetID( str ));
		_id = QString::fromStdString(str);
	}
	return _id;
}

void Cam::setState(const State& state)
{
	_state = state;
}

State Cam::state() const
{
	return _state;
}

void Cam::open()
{
	if(_state != closed)
		return error("cant open " + stateToString(_state) + " cam");

	f(_vcam->Open( VmbAccessModeFull ),"open "+_name);
	//state change will be handled by CamObserver::CameraStateChanged > Core::onCameraChanged > Cam::onChanged
}

void Cam::close()
{
	if(_state != opened)
		return error("cant close " + stateToString(_state) + " cam");

	f(_vcam->Close(),"close cam #"+_id);
	//state change will be handled by CamObserver::CameraStateChanged > Core::onCameraChanged > Cam::onChanged
}

QStringList Cam::getStreamFeatures( QStringList features )
{
	if(_state != opened && _state != recording )
	{
		error("cant access features on " + stateToString(_state) + " cam");
		return {""};
	}

	StreamPtrVector streams;
	_vcam->GetStreams(streams);
	return getFeatures(streams.at(0), features);
}

QStringList Cam::getCamFeatures( QStringList features )
{
	if(_state != opened )
	{
		error("cant access features on " + stateToString(_state) + " cam");
		return {""};
	}

	return getFeatures(_vcam, features);
}

QStringList Cam::getFeatures( FeatureContainerPtr camOrStream, QStringList features)
{
	QStringList allFeaturesWithValue;
	for( auto feature : features )
	{
		try
		{
			QString featureWithValue = feature + ": ";
			FeaturePtr pFeature;
			g( camOrStream->GetFeatureByName(feature.toStdString().c_str(), pFeature) ); //, QString("GetFeatureByName: %1").arg(feature)

			VmbFeatureDataType type;
			g( pFeature->GetDataType(type) ); //"Feature->GetDataType"

			switch(type)
			{
				case VmbFeatureDataInt:
				{
					VmbInt64_t value;
					g( pFeature->GetValue(value) );
					featureWithValue += QString::number(value);
					break;
				}
				case VmbFeatureDataFloat:
				{
					double value;
					g( pFeature->GetValue(value) );
					featureWithValue += QString::number(value);
					break;
				}
				case VmbFeatureDataEnum:
				{
					std::string value;
					g( pFeature->GetValue(value) );
					featureWithValue += QString::fromStdString(value);
					break;
				}
				case VmbFeatureDataString:
				{
					std::string value;
					g( pFeature->GetValue(value) );
					featureWithValue += QString::fromStdString(value);
					break;
				}
				case VmbFeatureDataBool:
				{
					bool value;
					g( pFeature->GetValue(value) );
					featureWithValue += QString::number(value);
					break;
				}
				default:
					qDebug() << "Unsupported feature type for " << feature;
					continue;

			}
			allFeaturesWithValue << featureWithValue;
		}
		catch (const std::runtime_error& xx)
		{
			qDebug() << __FUNCTION__ << "Exception caught: " << xx.what();
			// break;
		}
	}
	print(allFeaturesWithValue);
	return allFeaturesWithValue;
}

// id,name,model,serial,interfaceID,state
QString Cam::camInfo()
{
	QString camInfo = _name + DELIM + _ip + DELIM;

	if( state() == Cam::disconnected )
		return camInfo + "disconnected";

	try
	{
		camInfo += id() + DELIM;

		std::string str;
		g(_vcam->GetModel( str ));
		camInfo += QString::fromStdString(str) + DELIM;

		g(_vcam->GetSerialNumber( str ));
		camInfo += QString::fromStdString(str) + DELIM;

		g(_vcam->GetInterfaceID( str ));
		camInfo += QString::fromStdString(str) + DELIM;

		camInfo += stateToString(_state) + DELIM;
		if( _state == recording)
			camInfo += _rec->progressBar() + DELIM;
		else
			camInfo += "recdur: " + formatTime( _rec->dur().count() * 1000 ) + DELIM;

	}
	catch(const std::exception& e)
	{
		qDebug() << __LINE__ << "-" << __PRETTY_FUNCTION__ << "exception:" << e.what();
	}

	return camInfo;
}

QString Cam::stateToString(const State& state)
{
	switch (state)
	{
		case disconnected:
			return "disconnected";
		case closed:
			return "closed";
		case opened:
			return "opened";
		case recording:
			return "recording";
		default:
			return "unknown";
	}
}

void Cam::onChanged(UpdateTriggerType type)
{
	switch (type)
	{
		case UpdateTriggerType::UpdateTriggerPluggedIn:
			info(name()+" connected");
			setState(closed);
			//_vcam ptr set in core
			break;
		case UpdateTriggerType::UpdateTriggerPluggedOut:
			info(name()+" disconnected");
			setState(disconnected);
			_vcam.reset();
			break;
		case UpdateTriggerType::UpdateTriggerOpenStateChanged:

			// qDebug() << __LINE__ << "-" << __PRETTY_FUNCTION__ << "state changed" << stateToString(_state);
			if( _state == opened )
				_state = closed;
			else if ( _state == closed )
				_state = opened;
			break;
		default:
			error("onChanged: unknown trigger type");
	}
}
//write out settings.xml for current open cam
void Cam::saveSettings()
{
	if(_state != opened)
		return error("can only save settings on opened cam. (its " + stateToString(_state) + ")");

	VmbFeaturePersistSettings_t settingsStruct;
	settingsStruct.loggingLevel = VmbLogLevelWarn; //err/warn/dbg
	settingsStruct.maxIterations = 8;
	settingsStruct.persistType = VmbFeaturePersistStreamable; //dont store IP address!
	settingsStruct.modulePersistFlags = VmbModulePersistFlagsType::VmbModulePersistFlagsNone; //


	QString file="settings_"+name()+".xml";
	if (f(_vcam->SaveSettings( file.toStdString().c_str(), &settingsStruct )))
		error( "Could not save camera settings to '" + file + "'" );
	else
		info("Saved settings to " + file);
}

//load settings.xml for current open cam
void Cam::loadSettings()
{
	using utils::settingsFile;

	if(_state != opened)
		return error("can only save settings on opened cam. (its " + stateToString(_state) + ")");

	if(f(_vcam->LoadSettings( settingsFile().toStdString().c_str())))
		error( "Could not load camera settings from '" + settingsFile() + "'" );
	else
		info("Loaded settings from " + settingsFile());
}
