#include "console.h"
#include "qcoreapplication.h"
#include "qnamespace.h"

#include "../utils.h"
#include "qset.h"

#include <QDebug>
#include <QCoreApplication>
#include <QTimer>

#include <iostream>
#include <cmath>

using utils::DELIM;

Console::Console(): QObject() {}; // Private constructor to prevent instantiation

Console* Console::getInstance()
{
	static Console* instance = new Console; // Static local variable ensures single instance
	return instance;
}

void Console::listenKeys()
{
	// listen to keyevents in endless loop
	/* This is an expensive or blocking operation, why we use threaded approach to not block main event loop! */
	std::string input;
	while( utils::running )
	{
			std::getline(std::cin, input); //blox
			if( input.empty() )
				continue;
			else if ( input.find_first_not_of("0123456789") == std::string::npos)
			{
				emit numberEntered(std::stoi(input)); // Number
			}
			else if (input.size() >= 1)
			{
				emit keyPressed(input[0]); // Single char (excluding single digits)
				if( input[0] == 'q' )
					utils::running = false;
			}
			else
			{
				emit lineEntered(QString::fromStdString(input)); // Line
			}
		}
		// qDebug() << __LINE__ << "-" << __PRETTY_FUNCTION__ << "";
}

void Console::error(const int& errCode )
{
	if( errCode )
	{
		QString errStr = utils::errorCodeToMessage( errCode );
		qDebug().noquote() << errStr;
	}
}

void Console::error(const QString& str, const int& errCode=0 )
{
	if( str != "" )
		// qDebug().noquote() << "💩 " << str;
		// qDebug().noquote() << "💣 " << str;
		qDebug().noquote() << "💥 " << str;
		// qDebug().noquote() << "🔥 " << str;
	error(errCode);
}

void Console::warning(const QString& str )
{
	if( str != "" )
		qDebug().noquote() << "⚠️ " << str;
}

void Console::print(const QString& str )
{
	qDebug().noquote() << str;
}

void Console::print(const QStringList& list )
{

	for ( auto str : list  )
	{
		str.replace(DELIM," ");
		qDebug().noquote() << str << " ";
	}
}

void Console::printVersion()
{
	auto version = utils::getVersions();
	print(version);
}

void Console::printHelp()
{
	//sa Controller::keyPress
	QStringList str;
	str << "HELP!"
	<< "-------------------------------"
	<< "h: print this help"
	<< "1-9: select camera"
	<< "d/D: change record duration of current/all camera(s)"
	<< "o: open current camera"
	<< "c: close current camera"
	<< "r: start recording"
	<< "t: stop recording"
	<< "s: show recording stats"
	<< "i: show info"
	<< "l: list available cameras"
	<< "k: deteKt cameras"
	<< "y: load cam settings"
	<< "u: save cam settings"
	<< "v: print versions"
	<< "q: quit"
	// XXX add new func
	<< "-------------------------------";

	print(str);
}

// ###############################
// Controller
Controller::Controller() :
getDur1(false),
getDurAll(false)
{
	auto console = Console::getInstance();
	console->moveToThread(&thread);
	connect(&thread, &QThread::finished, console, &QObject::deleteLater);
	connect(this, &Controller::operate, console, &Console::listenKeys);

	connect(console, &Console::keyPressed, this, &Controller::onKeyPressed);
	connect(console, &Console::lineEntered, this, &Controller::onLineEntered);
	connect(console, &Console::numberEntered, this, &Controller::onNumberEntered);

	QTimer::singleShot(0, this, SIGNAL(operate()));
	thread.start();
}

Controller::~Controller()
{
	// qDebug() << __LINE__ << "-" << __PRETTY_FUNCTION__ << "";
	thread.quit();
	thread.wait(); //wait for thread to finish
	qDebug().noquote() << __FUNCTION__ << "():" << __LINE__ << ": ";
}

void Controller::onKeyPressed(const QChar& key)
{
	auto console = Console::getInstance();
	// qDebug().noquote() << __FUNCTION__ << ": " << key;

	switch (key.unicode())
	{
		// XXX add new func
		case 'd':
			console->print("Enter duration in seconds: ");
			getDur1 = true;
			break;

		case 'D':
			console->print("Enter duration in seconds: ");
			getDurAll = true;
			break;

		case 'c':
			emit closeCam();
			break;

		case 'o':
			emit openCam();
			break;

		case 'l':
			emit listCams();
			break;

		case 'k':
			emit detectCams();
			break;

		case 'r':
			emit startRecording();
			break;

		case 't':
			emit stopRecording();
			break;

		case 's':
			emit showRecordingStats();
			break;

		case 'i':
			emit showInfo();
			break;

		case 'y':
			emit loadSettings();
			break;

		case 'u':
			emit saveSettings();
			break;

		case 'v':
			console->printVersion();
			break;

		case 'q':
			qApp->quit();
			break;

		default:
			console->printHelp();
			break;
	}
}

void Controller::onLineEntered(const QString& line)
{
	qDebug().noquote() << "Line entered: " << line;
}

void Controller::onNumberEntered(int number)
{
	// qDebug().noquote() << "Number entered: " << number;

	if( getDur1 )
	{
		getDur1 = false;
		emit setDuration1((seconds)number);
	}
	else if( getDurAll )
	{
		getDurAll = false;
		emit setDurationAll((seconds)number);
	}
	else
	{
		emit cam(number-1);
	}
}
