#include "record.h"
#include "opencv2/highgui.hpp"
#include "qjsonvalue.h"
#include "utils.h"
#include "cam.h"
#include "frameobserver.h"

#include <QDebug>
#include <QTimer>
#include <QStorageInfo>
#include <chrono>

using namespace VmbCPP;
using namespace std::chrono;

Record::Record() : IPrinter(),
_cam(nullptr),
_timer(new QTimer(this)),
_updateTimer(new QTimer(this)),
_dir(),
_payloadSize(0LL),
_pixelFormat(""),
_frames(FramePtrVector(utils::threadsPerCam()))
{
	_timer->setSingleShot(true);
	_timer->setInterval( duration_cast<milliseconds>(utils::recDuration()) );
	connect(_timer, SIGNAL(timeout()), this, SLOT(stop()));


	_updateTimer->setInterval(duration_cast<milliseconds>(updateInterval));
	connect(_updateTimer, SIGNAL(timeout()), this, SLOT(showProgressBar()));
}

void Record::init(CamPtr cam)
{
	_cam = cam;
}

void Record::setDur(seconds dur)
{
	// qDebug() << __LINE__ << "-" << __PRETTY_FUNCTION__ << "";
	if( _cam->state() == Cam::recording )
		return error("cant change dur, while cam is recording");

	_timer->setInterval(dur);
}

seconds Record::dur()
{
	return duration_cast<seconds>(_timer->intervalAsDuration());
}

/* Does:
	get Payloadsize
	get Pixelformat - eg RGB8
	Allocate memory for frame buffer
	Register frame observer / callback for each frame
	Announce frame to the API
	Start the capture engine (API)
	Put frame into the frame queue
	Start the acquisition engine ( _curCam)

	2d get frameID
*/
void Record::start()
{
	//create dir for recording
	if(!checkDir(utils::outDir() + QDir::separator() + _cam->name() + QDir::separator() + curDateTime()))
		return;

	try
	{
		auto vcam = _cam->getCameraPtr();
		FeaturePtr pFeature;

		g( vcam->GetFeatureByName ("PayloadSize", pFeature )); //xxx adjust: set to 7062528
		g( pFeature->GetValue (_payloadSize));
		print("PayloadSize",_payloadSize);

		g( vcam->GetFeatureByName("PixelFormat", pFeature));
		g( pFeature->GetValue(_pixelFormat));
		print("PixelFormat",_pixelFormat);

		for( auto& frame : _frames )
		{
			frame.reset(new Frame(_payloadSize));
			g( frame->RegisterObserver( IFrameObserverPtr( new FrameObserver(_cam) )),"register frame observer");
			g( vcam->AnnounceFrame(frame), "announce frame");
		}

		g( vcam->StartCapture(), "start capture");

		for( auto& frame : _frames )
			g( vcam->QueueFrame(frame),"QueueFrame");

		g( vcam->GetFeatureByName ("AcquisitionStart", pFeature ),"get AcquisitionStart");
		g( pFeature->RunCommand(),"run AcquisitionStart");

	}
	catch (const std::runtime_error& xx)
	{
		qDebug() << __FUNCTION__ << "Exception caught: " << xx.what();
		return;
	}

	_cam->setState(Cam::recording);
	_timer->start();
	_updateTimer->start();
}

/* Does:
	Stop the capture engine (API) + discard all pending callbacks
	Flush the frame queue
	Revoke all frames from the API
	Unregister the frame observer / callback
*/
void Record::stop()
{
	showProgressBar();

	_updateTimer->stop();
	if(_timer->isActive())
		_timer->stop(); //user stop or error

	try
	{
		FeaturePtr pFeature;
		auto vcam = _cam->getCameraPtr();

		g(vcam->GetFeatureByName ("AcquisitionStop", pFeature ),"get AcquisitionStop");
		g(pFeature->RunCommand(),"run AcquisitionStop");
		g(vcam->EndCapture(),"end capture");
		g(vcam->FlushQueue(),"flush queue");
		g(vcam->RevokeAllFrames(),"revoke all frames");
		for( FramePtr frame : _frames )
			g(frame->UnregisterObserver(),"unregister observer");

		_cam->setState(Cam::opened);
	}
	catch (const std::runtime_error& xx)
	{
		qDebug() << __FUNCTION__ << "Exception caught: " << xx.what();
		return;
	}

	auto STATS = stats();
	info(STATS);
	write(STATS,"stats");
}

bool Record::write(QStringList sl, QString fileName)
{

	QDir dir(_dir);
	QString filePath = dir.filePath(fileName); // Construct the file path
	QFile file(filePath);
	if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) {
		error("Failed to open file:"+filePath);
		return false;
	}

	QTextStream out(&file);
	out << sl.join("\n");
	return true;
} // calls file.close();


QStringList Record::stats()
{
	QStringList _stats;
	_stats << "-------------------------------";
	_stats << "cam features:";
	_stats << _cam->getCamFeatures({
		"AcquisitionFrameRate",
		"ExposureTime",
		"Gain",
		"SensorWidth",
		"SensorHeight",
		"DeviceTemperature",
		"PayloadSize",
		"PixelFormat",
		"PixelSize",
	});

	_stats << "-------------------------------";
	_stats << "stream stats:";
	_stats << _cam->getStreamFeatures({
		"StatFrameRate",
		"StatFrameDropped",
		"StatFrameRescued",
		"StatFrameShoved",
		"StatFrameUnderrun",
		"StatPacketErrors",
		"StatPacketMissed",
		"StatPacketReceived",
		"StatPacketRequested",
		"StatPacketResent",
		"StatTimeElapsed",
		"StatFrameDelivered",
	});

	return _stats;
}

void Record::showProgressBar()
{
	info( progressBar() );
}

QString Record::progressBar(int percentage)
{
	if(percentage == -1)
		return IPrinter::progressBar(_timer);
	else
		return IPrinter::progressBar(percentage);
}

void Record::checkDiskSpace(const QString &dir, seconds dur)
{
	QStorageInfo storageInfo(dir);

	// In bytes
	qint64 available = storageInfo.bytesAvailable();
	qint64 total = storageInfo.bytesTotal();
	qint64 used = calcDirSize(dir);

	int usedPercentage = (double(used) / total) * 100;
	progressBar(usedPercentage); // << cuz of this line, the function is here not in utils...

	qint64 estimated = dur.count() * utils::fps() * utils::frameSize();

	auto formatSize = [](qint64 size) -> QString {
		double displaySize = size;
		QString unit = "MB";

		const qint64 ONE_MB = 1024LL * 1024LL;
		const qint64 ONE_GB = ONE_MB * 1024LL;
		const qint64 ONE_TB = ONE_GB * 1024LL;

		if (size >= ONE_TB) {
			displaySize = size / double(ONE_TB);
			unit = "TB";
		} else if (size >= ONE_GB) {
			displaySize = size / double(ONE_GB);
			unit = "GB";
		} else {
			displaySize = size / double(ONE_MB);
		}

		return QString::number(displaySize, 'f', 2) + " " + unit;
	};

	print("Used Space",formatSize(used));
	print("Available Space",formatSize(available));
	print("Total Space",formatSize(total));
	print(QString("Used Percentage: %1%").arg(usedPercentage));
	print("Estimated Recording Space",formatSize(estimated));

	// Check if estimated recording space exceeds available space
	if (estimated > available) {
		warning("Estimated recording space exceeds available space!");
	}
}


//calc basedir size
qint64 Record::calcDirSize(const QDir &dirrr) {
	QDir dir(dirrr);
	qint64 size = 0;

	// Recursively calculate size of files in the directory
	foreach(QFileInfo fileInfo, dir.entryInfoList(QDir::Files | QDir::Dirs | QDir::NoDotAndDotDot, QDir::DirsFirst))
	{
		if (fileInfo.isDir())
			size += calcDirSize(fileInfo.absoluteFilePath());
		else
			size += fileInfo.size();
	}

	return size;
}

// check if dir exists and if it doesnt, create it
// set _dir
// checkdiskspace()
bool Record::checkDir(QString dirPath)
{
	if( dirPath.isEmpty() )
		dirPath = utils::outDir();

	QDir dir(dirPath);
	if (!dir.exists() && !dir.mkpath(".") )
	{
		error("could not create directory: " + dirPath);
		return false;
	}

	_dir = dirPath;

	checkDiskSpace(dirPath, dur());
	info("writing frames to" + dirPath);
	return true;
}
