#include "utils.h"

#include "opencv2/core/version.hpp"

#include <random>
#include <iostream>

#include "VmbCPP/VmbCPP.h"
#include "qglobal.h"
#include <VmbC/VmbCommonTypes.h>

#include <QDir>
#include <QThread>
#include <QCoreApplication>
#include <QFile>
#include <QJsonDocument>
#include <QJsonArray>
#include <QJsonObject>
#include <QFileInfoList>


using namespace VmbCPP;

namespace utils
{

const int APP_VERSION_MAJOR = 0;
const int APP_VERSION_MINOR = 4;
const int APP_VERSION_PATCH = 0;

std::atomic<bool> running(true);

int ncam(const int& n)
{
	static const int _ncam = n;

#ifdef DEBUG
	return 2;
#endif
	return _ncam;
}

int threadsPerCam()
{
#ifdef DEBUG
	return 8;
#else
	return std::ceil(static_cast<double>(QThread::idealThreadCount()-2) / ncam() ); //round up! -2 for main and console
#endif
}

qint64 frameSize( qint64 frameSize )
{
	static qint64 _frameSize;
	if( frameSize != 0 )
		_frameSize = frameSize;
	return _frameSize == 0 ? 1024*1024 : _frameSize; //1mb dflt
}

int fps( int fps )
{
	static int _fps;
	if( fps != 0 )
		_fps = fps;
	return _fps == 0 ? 17 : _fps; //dflt
}

seconds recDuration(seconds dur)
{
	static seconds _recDuration;
	if( dur != 0s )
		_recDuration = dur;

	return _recDuration == 0s ? 10s : _recDuration; //dflt
}

//general basedir
QString outDir(QString dirname)
{
	static QString _outDir;
	if( _outDir.isEmpty() )
	{
		if( dirname.isEmpty() )
			_outDir = QCoreApplication::applicationDirPath();
		else
			_outDir = dirname;
	}

	return _outDir;
}


const QString getVersion()
{
	return QString("# Recorder: %1.%2.%3").arg(APP_VERSION_MAJOR).arg(APP_VERSION_MINOR).arg(APP_VERSION_PATCH);
}
const QStringList getVersions()
{
	VmbVersionInfo_t version;
	VmbSystem::GetInstance().QueryVersion(version);

	QStringList s;
	s << "### Versions";
	s << QString("# Recorder: %1.%2.%3").arg(APP_VERSION_MAJOR).arg(APP_VERSION_MINOR).arg(APP_VERSION_PATCH);
	s << QString("#    Vimba: %1.%2.%3").arg(version.major).arg(version.minor).arg(version.patch);
	s << QString("#       Qt: %1.%2.%3").arg(QT_VERSION_MAJOR).arg(QT_VERSION_MINOR).arg(QT_VERSION_PATCH);
	s << QString("#   openCV: %1.%2.%3").arg(CV_VERSION_MAJOR).arg(CV_VERSION_MINOR).arg(CV_VERSION_REVISION);

	return s;
}

const QString errorCodeToMessage( VmbError_t err )
{
	QString msg = "💣 ";
	switch( err )
	{
		default:                        msg += "Unknown"; break;
		case VmbErrorJustInfo:          msg = ""; break; //custom
		case VmbErrorSuccess:           msg = "success! 😎"; break; //
		case VmbErrorInternalFault:     msg += "Unexpected fault in VmbApi or driver."; break;
		case VmbErrorApiNotStarted:     msg += "API not started."; break;
		case VmbErrorNotFound:          msg += "Not found."; break;
		case VmbErrorBadHandle:         msg += "Invalid handle "; break;
		case VmbErrorDeviceNotOpen:     msg += "Device not open."; break;
		case VmbErrorInvalidAccess:     msg += "Invalid access."; break;
		case VmbErrorBadParameter:      msg += "Bad parameter."; break;
		case VmbErrorStructSize:        msg += "Wrong DLL version."; break;
		case VmbErrorMoreData:          msg += "More data returned than memory provided."; break;
		case VmbErrorWrongType:         msg += "Wrong type."; break;
		case VmbErrorInvalidValue:      msg += "Invalid value."; break;
		case VmbErrorTimeout:           msg += "Timeout."; break;
		case VmbErrorOther:             msg += "TL error."; break;
		case VmbErrorResources:         msg += "Resource not available."; break;
		case VmbErrorInvalidCall:       msg += "Invalid call."; break;
		case VmbErrorNoTL:              msg += "TL not loaded."; break;
		case VmbErrorNotImplemented:    msg += "Not implemented."; break;
		case VmbErrorNotSupported:      msg += "Not supported."; break;
			// ------------------------------- own errCodes -------------------------------
		case errWarning:                msg = "⚠️ "; break;
		case errCamIdx:                 msg += "Camera index out of range."; break;
		case errNoCams:                 msg += "no cameras found."; break;
		case errRegisterCamObserver:    msg += "cant register camobserver."; break;
		case errLessCamsThanExpected:   msg += "less cameras found, than expected"; break;
		case errRecordInvalid:          msg += "Record invalid"; break;

		// ... add more
	}
	return msg;
}

qint64 parseInt(const QString& key, const QJsonObject& jsonObject)
{
	if (!jsonObject.contains(key) || !jsonObject[key].isDouble())
	{
		qWarning() << "Missing or invalid " << key;
		return 0;
	}
	return jsonObject[key].toInt();
}

QString parseString(const QString& key, const QJsonObject& jsonObject)
{
	if (!jsonObject.contains(key) || !jsonObject[key].isString())
	{
		qWarning() << "Missing or invalid " << key;
		return QString();
	}
	return jsonObject[key].toString();
}

bool parseBool(const QString& key, const QJsonObject& jsonObject)
{
	if (!jsonObject.contains(key) || !jsonObject[key].isBool())
	{
		qWarning() << "Missing or invalid " << key;
		return false;
	}
	return jsonObject[key].toBool();
}

bool parseConfig(QList<QPair<QString,QString>>& parsedCameras)
{
	// qDebug() << "open " << configFile();
	QFile file(configFile());
	if (!file.open(QIODevice::ReadOnly))
	{
		qWarning() << "Failed. Could not open config: " + configFile();
		return false;
	}

	QByteArray jsonData = file.readAll();
	file.close();

	QJsonParseError parseError;
	QJsonDocument doc = QJsonDocument::fromJson(jsonData, &parseError);
	if (parseError.error != QJsonParseError::NoError)
	{
		qWarning() << "Error parsing JSON:" << parseError.errorString();
		return false;
	}

	if (!doc.isObject())
	{
		qWarning() << "JSON content is not an object";
		return false;
	}

	QJsonObject jsonObject = doc.object();

	// parse globals
	int dur = parseInt("recordDurationSeconds", jsonObject);
	recDuration( 1s * dur ); //convert to s
	fps( parseInt("fps", jsonObject) );
	frameSize( parseInt("frameSizeBytes", jsonObject));
	outDir( parseString("outDir", jsonObject) );

	// parse cameras
	if (!jsonObject.contains("cameras") || !jsonObject["cameras"].isArray())
	{
		qWarning() << "Missing or invalid cameras array";
		return false;
	}

	parsedCameras.clear();
	QJsonArray camerasArray = jsonObject["cameras"].toArray();
	for (const QJsonValue& value : camerasArray)
	{
		if (!value.isObject())
		{
				qWarning() << "Invalid camera entry:" << value;
				continue;
		}

		QJsonObject cameraObject = value.toObject();
		auto name = cameraObject["name"].toString();
		auto ip = cameraObject["ip"].toString();

		parsedCameras.append(qMakePair(name, ip));
	}
	ncam(parsedCameras.size());
	return true;
}

// needed to find cams via ip
QString configFile(QString filename)
{
	static QString _configFile;
	if( _configFile.isEmpty() )
		_configFile = filename;

	return _configFile.isEmpty() ? utils::getFirstFileInAppDirEndingWith("json") : _configFile;
}

QString getFirstFileInAppDirEndingWith( QString const& suffix )
{
	QDir dir(QCoreApplication::applicationDirPath());
	QStringList files = dir.entryList(QStringList() << "*."+suffix, QDir::Files);
	// QStringList const folders = source.entryList(QDir::NoDot | QDir::NoDotDot | QDir::Dirs);
	auto file =  files.first();
	return files.isEmpty() || suffix.isEmpty() ? QString() : file;
}

// optional - only set cam settings when they changed
QString settingsFile(QString filename)
{
	static QString _settingsFile;
	if( _settingsFile.isEmpty() )
		_settingsFile = filename;

	return _settingsFile.isEmpty() ? utils::getFirstFileInAppDirEndingWith("xml") : _settingsFile;
}

} // namespace utils
